//
// This file is part of Adguard for iOS (https://github.com/AdguardTeam/AdguardForiOS).
// Copyright © Adguard Software Limited. All rights reserved.
//
// Adguard for iOS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Adguard for iOS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Adguard for iOS. If not, see <http://www.gnu.org/licenses/>.
//

import Foundation
import UIKit

extension NSMutableAttributedString {
    enum AttachmentSize {
        case defaultSize
        case customSize(width: CGFloat, height: CGFloat)
    }

    struct AttachmentSettings {
        let image: UIImage
        let topEdge: CGFloat
        let leftEdge: CGFloat
        let size: AttachmentSize
    }

    func alignCenter(){
        let paragraph = NSMutableParagraphStyle()
        paragraph.alignment = .center

        let attributes: [NSAttributedString.Key : Any] = [NSAttributedString.Key.paragraphStyle: paragraph]
        self.addAttributes(attributes, range: NSRange(location: 0, length: length))
    }

    static func fromHtml(_ html: String, fontSize: CGFloat, color: UIColor, attachmentSettings: AttachmentSettings? = nil, textAlignment: NSTextAlignment = .left, lineBreakMode: NSLineBreakMode = .byWordWrapping) -> NSMutableAttributedString? {

        let style = NSMutableParagraphStyle()
        style.alignment = textAlignment
        style.lineBreakMode = lineBreakMode

        let format = NSMutableString(string: html)
        let wrapped = "<span style=\"font-family: -apple-system; font-size: \(fontSize); color: \(color.hex())\">\(format)</span>"
        guard let htmlData = wrapped.data(using: .utf8) else { return nil}

        guard let resultText = try? NSMutableAttributedString(
                data: htmlData,
                options: [.documentType: NSAttributedString.DocumentType.html,
                          .characterEncoding:NSNumber(value:String.Encoding.utf8.rawValue)],
                documentAttributes: nil) else { return nil }

        if let settings = attachmentSettings {

            let imageRange = resultText.mutableString.range(of: "%@")
            resultText.replaceCharacters(in: imageRange, with: "")

            let attachment = NSTextAttachment()
            attachment.image = settings.image

            let x = settings.leftEdge
            let y = -settings.topEdge
            let width: CGFloat
            let height: CGFloat
            switch settings.size {
            case .defaultSize:
                width = CGFloat(settings.image.cgImage!.width) / 2.5
                height = CGFloat(settings.image.cgImage!.height) / 2.5
            case .customSize(let w, let h):
                width = w
                height = h
            }


            attachment.bounds = CGRect(x: x, y: y, width: width, height: height)

            let attachmentString = NSAttributedString(attachment: attachment)
            resultText.insert(attachmentString, at: imageRange.location)
        }

        resultText.addAttributes([NSAttributedString.Key.paragraphStyle : style], range: NSRange(location: 0, length: resultText.length))

        return resultText
    }
}

extension String {

    func dataFromHex()-> Data {
        var hex = self
        var data = Data()
        while(hex.count > 0) {
            let subIndex = hex.index(hex.startIndex, offsetBy: 2)
            let c = String(hex[..<subIndex])
            hex = String(hex[subIndex...])
            var ch: UInt32 = 0
            Scanner(string: c).scanHexInt32(&ch)
            var char = UInt8(ch)
            data.append(&char, count: 1)
        }
        return data
    }

    static func randomString(length: Int) -> String {
        let letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
        return String((0..<length).map{ _ in letters.randomElement()! })
    }

    /// Returns tuple with attributed string that highlight passed occurancies and `found` flag
    /// It is true if any occurancies were found
    func highlight(occuranciesOf strings: Set<String>) -> (attrString: NSAttributedString, matchesFound: Bool) {
        let attributedString = NSMutableAttributedString(string: self)
        attributedString.addAttribute(NSAttributedString.Key.backgroundColor, value: UIColor.clear, range: NSRange(location: 0, length: self.count))
        guard !strings.isEmpty else { return (attributedString, false) }

        let highlightColor = UIColor.AdGuardColor.lightGreen1
        var foundCount = 0
        for string in strings {
            let nsRanges = ranges(of: string, options: [.caseInsensitive])
            nsRanges.forEach {
                foundCount += 1
                attributedString.addAttribute(NSAttributedString.Key.backgroundColor, value: highlightColor, range: $0)
            }
        }
        return (attributedString, foundCount != 0)
    }

    /// Returns NSRanges of all occurancies of substring in string
    func ranges(of substring: String, options: CompareOptions = [], locale: Locale? = nil) -> [NSRange] {
        var searchRange: Range<Index>?
        var result = [NSRange]()
        while let range = range(of: substring, options: options, range: (searchRange?.upperBound ?? startIndex)..<endIndex, locale: locale) {
            searchRange = range
            result.append(self.nsRange(from: range))
        }
        return result
    }

    var clearAttrString: NSAttributedString {
        let attributedString = NSMutableAttributedString(string: self)
        attributedString.addAttribute(NSAttributedString.Key.backgroundColor, value: UIColor.clear, range: NSRange(location: 0, length: self.count))
        return attributedString
    }

    // Adds space every 3 symbols
    // Use for formatting numbers
    static func formatSringNumber(number: Int) -> String{
        let stringNumber = String(number)
        var str = String(stringNumber.reversed())
        str = String(str.enumerated().map { $0 > 0 && $0 % 3 == 0 ? [",", $1] : [$1]}.joined())
        str = String(str.reversed())
        return str
    }

    func isValidUpstream() -> Bool {
        let charSet = CharacterSet(charactersIn: self)
        var validCharsSet = CharacterSet.urlPathAllowed
        validCharsSet.insert("%")
        validCharsSet.insert("[")
        validCharsSet.insert("]")
        return charSet.isSubset(of: validCharsSet) && self.count > 0
    }

    /*
     Discards port from IP address
     94.140.14.15:52 -> 94.140.14.15
     [2a10:50c0::bad1:ff]:53 -> 2a10:50c0::bad1:ff
     Returns initial string if fails to fetch IP address
     */
    func discardPortFromIpAddress() -> String {
        /// Check if string looks like ip address
        guard let firstChar = self.first, firstChar == "[" || firstChar.isNumber else {
            return self
        }

        if self.contains("[") && self.contains("]") {
            let leftQuotePosition = self.firstIndex(of: "[")!
            let rightQuotePosition = self.firstIndex(of: "]")!

            let leftQuoteIndex = self.index(leftQuotePosition, offsetBy: 1)
            let rightQuoteIndex = self.index(rightQuotePosition, offsetBy: -1)

            guard leftQuoteIndex < rightQuoteIndex else {
                return self
            }
            let ipAddress = self[leftQuoteIndex...rightQuoteIndex]
            return String(ipAddress)
        } else {
            let parts = self.split(separator: ":")
            guard parts.count == 2 else {
                return self
            }
            return String(parts[0])
        }
    }

    static func localizedString(_ key: String)->String {
        return ACLocalizedString(key, nil)
    }

    static func simpleSecondsFormatter(_ ms: NSNumber) -> String {
        let formatter = NumberFormatter()
        formatter.numberStyle = .decimal
        formatter.locale = .current
        formatter.minimumFractionDigits = 0

        let seconds = ms.doubleValue / 1000
        if seconds >= 1 {
            formatter.maximumFractionDigits = 1
            let formatterString = formatter.string(from: NSNumber(floatLiteral: seconds)) ?? "\(ms.intValue)"
            return String(format: String.localizedString("s_unit"), formatterString)
        }

        if ms.intValue >= 1 {
            formatter.maximumFractionDigits = 0
            let formatterString = formatter.string(from: NSNumber(floatLiteral: ms.doubleValue)) ?? "\(ms.intValue)"
            return String(format: String.localizedString("ms_unit"), formatterString)
        }

        formatter.maximumFractionDigits = 0
        let formatterString = formatter.string(from: ms) ?? "\(ms.intValue)"
        return String(format: String.localizedString("ms_unit"), formatterString)
    }

    /*
     Formats a number, devides thousands with separator, depending on current locale
     */
    static func simpleThousandsFormatting(_ number: NSNumber) -> String {
        let formatter = NumberFormatter()
        formatter.numberStyle = .decimal
        formatter.locale = .current
        formatter.minimumFractionDigits = 0
        formatter.maximumFractionDigits = 0
        return formatter.string(from: number) ?? "\(number.intValue)"
    }

    /**
    Converts number to string, ignores decimal part
    Formatted by locale
    */
    static func formatNumberByLocale(_ number: NSNumber) -> String {
        let formatter = NumberFormatter()
        formatter.numberStyle = .decimal
        formatter.locale = .current
        formatter.minimumFractionDigits = 0
        formatter.maximumFractionDigits = 0

        let decimalNumber: Double = number.doubleValue

        /* Asian countries from this array have a different way of formatting numbers */
        let specialAsianCountriesCodes = ["zh", "ko", "ja"]
        let currentLocaleLanguageCode = Locale.current.languageCode ?? ""

        if specialAsianCountriesCodes.contains(currentLocaleLanguageCode) {

            let hundredMillions = decimalNumber / 100000000
            if hundredMillions > 1 {
                let hundredMillionsString = formatter.string(from: NSNumber(floatLiteral: hundredMillions)) ?? "0"
                return String(format: String.localizedString("hundred_millions_unit"), hundredMillionsString)
            }

            let tenThousands = decimalNumber / 10000
            if tenThousands > 100 {
                let tenThousandsString = formatter.string(from: NSNumber(floatLiteral: tenThousands)) ?? "0"
                return String(format: String.localizedString("ten_thousands_unit"), tenThousandsString)
            }

            return formatter.string(from: number) ?? "0"
        }

        /* This is normal formatting */
        let millions = decimalNumber / 1000000
        if millions > 1 {
            let millionsString = formatter.string(from: NSNumber(floatLiteral: millions)) ?? "0"
            return String(format: String.localizedString("millions_unit"), millionsString)
        }

        let thousands = decimalNumber / 1000

        if thousands > 100 {
            formatter.maximumFractionDigits = 0
            let thousandsString = formatter.string(from: NSNumber(floatLiteral: thousands)) ?? "0"
            return String(format: String.localizedString("thousands_unit"), thousandsString)
        }

        if thousands > 10 {
            formatter.maximumFractionDigits = 1
            let thousandsString = formatter.string(from: NSNumber(floatLiteral: thousands)) ?? "0"
            return String(format: String.localizedString("thousands_unit"), thousandsString)
        }

        if thousands > 1 {
            formatter.maximumFractionDigits = 2
            let thousandsString = formatter.string(from: NSNumber(floatLiteral: thousands)) ?? "0"
            return String(format: String.localizedString("thousands_unit"), thousandsString)
        }

        return formatter.string(from: number) ?? "0"
    }

    /**
    Generates subdomains from top level domain
    */
    static func generateSubDomains(from domain: String) -> [String] {
        let newDomain = domain.hasSuffix(".") ? String(domain.dropLast()) : domain

        var subdomains = newDomain.split(separator: ".")
        var domainsToReturn: [String] = []

        while subdomains.count >= 2{
            let newSubDomain = subdomains.joined(separator: ".")
            domainsToReturn.append(newSubDomain)
            subdomains = Array(subdomains.dropFirst())
        }
        return domainsToReturn
    }

    /* Checks if the string is valid email */
    func isValidEmail() -> Bool {
        let emailRegEx = "^(?!\\.)[A-Z0-9a-z._%+-]+@[A-Za-z0-9.-]+\\.[A-Za-z]{2,64}"
        let emailPred = NSPredicate(format:"SELF MATCHES %@", emailRegEx)
        return emailPred.evaluate(with: self)
    }

    func getQueryParametersFromQueryString() -> [String: String]? {
        var params: [String: String] = [:]
        let pairs = self.components(separatedBy: "&")
        for pair in pairs {
            let values = pair.components(separatedBy: "=")
            if values.count == 2 {
                params[values[0]] = values[1].removingPercentEncoding
            }

            if values.count == 1 {
                continue
            }
        }

        return params.count > 0 ? params : nil
    }



    /// Returns a substring after the first occurrence of `delimiter`.
    /// If the string does not contain the delimiter, returns the original string.
    func substringAfter(_ delimiter: Character) -> String { substringAfter(delimiter, self) }

    /// Returns a substring after the first occurrence of `delimiter`.
    /// If the string does not contain the delimiter, returns the original string.
    func substringAfter(_ delimiter: String) -> String { substringAfter(delimiter, self) }

    /// Returns a substring before the last occurrence of `delimiter`.
    /// If the string does not contain the delimiter, returns the original string.
    func substringBeforeLast(_ delimiter: Character) -> String { substringBeforeLast(delimiter, self) }

    /// Returns a substring before the last occurrence of `delimiter`.
    /// If the string does not contain the delimiter, returns the original string.
    func substringBeforeLast(_ delimiter: String) -> String { substringBeforeLast(delimiter, self) }

    /// Returns a substring after the first occurrence of `delimiter`.
    /// If the string does not contain the delimiter, returns `missingDelimiterValue`.
    func substringAfter(_ delimiter: Character, _ missingDelimiterValue: String) -> String {
        guard let first = firstIndex(of: delimiter),
              let afterDelimiterIndex = index(first, offsetBy: 1, limitedBy: endIndex)
        else { return missingDelimiterValue }

        return String(self[afterDelimiterIndex..<endIndex])
    }

    /// Returns a substring after the first occurrence of `delimiter`.
    /// If the string does not contain the delimiter, returns `missingDelimiterValue`.
    func substringAfter(_ delimiter: String, _ missingDelimiterValue: String) -> String {
        guard let range = range(of: delimiter) else { return missingDelimiterValue }
        
        return String(self[range.upperBound..<endIndex])
    }
    
    /// Returns a substring before the last occurrence of `delimiter`.
    /// If the string does not contain the delimiter, returns `missingDelimiterValue`.
    func substringBeforeLast(_ delimiter: Character, _ missingDelimiterValue: String) -> String {
        guard let last = lastIndex(of: delimiter) else { return missingDelimiterValue }

        return String(self[startIndex..<last])
    }

    /// Returns a substring before the last occurrence of `delimiter`.
    /// If the string does not contain the delimiter, returns `missingDelimiterValue`.
    func substringBeforeLast(_ delimiter: String, _ missingDelimiterValue: String) -> String {
        guard let range = range(of: delimiter, options: .backwards) else { return missingDelimiterValue }

        return String(self[startIndex..<range.lowerBound])
    }
}
